package sun.nio.cs;

import java.io.IOException;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Comparator;

public class CharsetMapping
{
    public static final char UNMAPPABLE_DECODING = '�';
    public static final int UNMAPPABLE_ENCODING = 65533;
    char[] b2cSB;
    char[] b2cDB1;
    char[] b2cDB2;
    int b2Min;
    int b2Max;
    int b1MinDB1;
    int b1MaxDB1;
    int b1MinDB2;
    int b1MaxDB2;
    int dbSegSize;
    char[] c2b;
    char[] c2bIndex;
    char[] b2cSupp;
    char[] c2bSupp;
    CharsetMapping.Entry[] b2cComp;
    CharsetMapping.Entry[] c2bComp;
    static Comparator<CharsetMapping.Entry> comparatorBytes;
    static Comparator<CharsetMapping.Entry> comparatorCP;
    static Comparator<CharsetMapping.Entry> comparatorComp;
    private static final int MAP_SINGLEBYTE = 1;
    private static final int MAP_DOUBLEBYTE1 = 2;
    private static final int MAP_DOUBLEBYTE2 = 3;
    private static final int MAP_SUPPLEMENT = 5;
    private static final int MAP_SUPPLEMENT_C2B = 6;
    private static final int MAP_COMPOSITE = 7;
    private static final int MAP_INDEXC2B = 8;
    int off;
    byte[] bb;
    
    public CharsetMapping() {
        this.off = 0;
    }
    
    public char decodeSingle(int n) {
        return this.b2cSB[n];
    }
    
    public char decodeDouble(int n, int n2) {
        if (n2 >= this.b2Min && n2 < this.b2Max) {
            n2 -= this.b2Min;
            if (n >= this.b1MinDB1 && n <= this.b1MaxDB1) {
                n -= this.b1MinDB1;
                return this.b2cDB1[n * this.dbSegSize + n2];
            }
            if (n >= this.b1MinDB2 && n <= this.b1MaxDB2) {
                n -= this.b1MinDB2;
                return this.b2cDB2[n * this.dbSegSize + n2];
            }
        }
        return '�';
    }
    
    public char[] decodeSurrogate(int n, char[] dst) {
        int toIndex = this.b2cSupp.length / 2;
        int binarySearch = Arrays.binarySearch(this.b2cSupp, 0, toIndex, (char)n);
        if (binarySearch >= 0) {
            Character.toChars(this.b2cSupp[toIndex + binarySearch] + 131072, dst, 0);
            return dst;
        }
        return null;
    }
    
    public char[] decodeComposite(CharsetMapping.Entry entry, char[] array) {
        int bytes = findBytes(this.b2cComp, entry);
        if (bytes >= 0) {
            array[0] = (char)this.b2cComp[bytes].cp;
            array[1] = (char)this.b2cComp[bytes].cp2;
            return array;
        }
        return null;
    }
    
    public int encodeChar(char c) {
        int n = this.c2bIndex[c >> 8];
        if (n == '\uffff') {
            return 65533;
        }
        return this.c2b[n + (c & 'ÿ')];
    }
    
    public int encodeSurrogate(char high, char low) {
        int codePoint = Character.toCodePoint(high, low);
        if (codePoint < 131072 || codePoint >= 196608) {
            return 65533;
        }
        int toIndex = this.c2bSupp.length / 2;
        int binarySearch = Arrays.binarySearch(this.c2bSupp, 0, toIndex, (char)codePoint);
        if (binarySearch >= 0) {
            return this.c2bSupp[toIndex + binarySearch];
        }
        return 65533;
    }
    
    public boolean isCompositeBase(CharsetMapping.Entry entry) {
        return entry.cp <= 12791 && entry.cp >= 230 && findCP(this.c2bComp, entry) >= 0;
    }
    
    public int encodeComposite(CharsetMapping.Entry entry) {
        int comp = findComp(this.c2bComp, entry);
        if (comp >= 0) {
            return this.c2bComp[comp].bs;
        }
        return 65533;
    }
    
    public static CharsetMapping get(final InputStream inputStream) {
        return (CharsetMapping)AccessController.doPrivileged(new PrivilegedAction<CharsetMapping>() {
            @Override
            public CharsetMapping run() {
                return new CharsetMapping().load(inputStream);
            }
        });
    }
    
    static int findBytes(CharsetMapping.Entry[] a, CharsetMapping.Entry key) {
        return Arrays.binarySearch(a, 0, a.length, key, CharsetMapping.comparatorBytes);
    }
    
    static int findCP(CharsetMapping.Entry[] a, CharsetMapping.Entry key) {
        return Arrays.binarySearch(a, 0, a.length, key, CharsetMapping.comparatorCP);
    }
    
    static int findComp(CharsetMapping.Entry[] a, CharsetMapping.Entry key) {
        return Arrays.binarySearch(a, 0, a.length, key, CharsetMapping.comparatorComp);
    }
    
    private static final boolean readNBytes(InputStream inputStream, byte[] b, int i) throws IOException {
        int off = 0;
        while (i > 0) {
            int read = inputStream.read(b, off, i);
            if (read == -1) {
                return false;
            }
            i -= read;
            off += read;
        }
        return true;
    }
    
    private char[] readCharArray() {
        int n = (this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF);
        char[] array = new char[n];
        for (int i = 0; i < n; ++i) {
            array[i] = (char)((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        }
        return array;
    }
    
    void readSINGLEBYTE() {
        char[] charArray = this.readCharArray();
        for (int i = 0; i < charArray.length; ++i) {
            int n = charArray[i];
            if (n != '�') {
                this.c2b[this.c2bIndex[n >> 8] + (n & 'ÿ')] = (char)i;
            }
        }
        this.b2cSB = charArray;
    }
    
    void readINDEXC2B() {
        char[] charArray = this.readCharArray();
        int i = charArray.length - 1;
        while (i >= 0) {
            if (this.c2b == null && charArray[i] != -1) {
                Arrays.fill(this.c2b = new char[charArray[i] + 256], '�');
                break;
            }
            --i;
        }
        this.c2bIndex = charArray;
    }
    
    char[] readDB(int n, int n2, int n3) {
        char[] charArray = this.readCharArray();
        for (int i = 0; i < charArray.length; ++i) {
            int n4 = charArray[i];
            if (n4 != '�') {
                this.c2b[this.c2bIndex[n4 >> 8] + (n4 & 'ÿ')] = (char)((i / n3 + n) * 256 + (i % n3 + n2));
            }
        }
        return charArray;
    }
    
    void readDOUBLEBYTE1() {
        this.b1MinDB1 = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.b1MaxDB1 = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.b2Min = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.b2Max = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.dbSegSize = this.b2Max - this.b2Min + 1;
        this.b2cDB1 = this.readDB(this.b1MinDB1, this.b2Min, this.dbSegSize);
    }
    
    void readDOUBLEBYTE2() {
        this.b1MinDB2 = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.b1MaxDB2 = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.b2Min = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.b2Max = ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF));
        this.dbSegSize = this.b2Max - this.b2Min + 1;
        this.b2cDB2 = this.readDB(this.b1MinDB2, this.b2Min, this.dbSegSize);
    }
    
    void readCOMPOSITE() {
        char[] charArray = this.readCharArray();
        int n = charArray.length / 3;
        this.b2cComp = new CharsetMapping.Entry[n];
        this.c2bComp = new CharsetMapping.Entry[n];
        int i = 0;
        int n2 = 0;
        while (i < n) {
            CharsetMapping.Entry o = new CharsetMapping.Entry();
            o.bs = charArray[n2++];
            o.cp = charArray[n2++];
            o.cp2 = charArray[n2++];
            this.b2cComp[i] = o;
            this.c2bComp[i] = o;
            ++i;
        }
        Arrays.sort(this.c2bComp, 0, this.c2bComp.length, CharsetMapping.comparatorComp);
    }
    
    CharsetMapping load(InputStream inputStream) {
        try {
            int n = (inputStream.read() & 0xFF) << 24 | (inputStream.read() & 0xFF) << 16 | (inputStream.read() & 0xFF) << 8 | (inputStream.read() & 0xFF);
            this.bb = new byte[n];
            this.off = 0;
            if (!readNBytes(inputStream, this.bb, n)) {
                throw new RuntimeException("Corrupted data file");
            }
            inputStream.close();
            while (this.off < n) {
                switch ((this.bb[this.off++] & 0xFF) << 8 | (this.bb[this.off++] & 0xFF)) {
                    case 8: {
                        this.readINDEXC2B();
                        continue;
                    }
                    case 1: {
                        this.readSINGLEBYTE();
                        continue;
                    }
                    case 2: {
                        this.readDOUBLEBYTE1();
                        continue;
                    }
                    case 3: {
                        this.readDOUBLEBYTE2();
                        continue;
                    }
                    case 5: {
                        this.b2cSupp = this.readCharArray();
                        continue;
                    }
                    case 6: {
                        this.c2bSupp = this.readCharArray();
                        continue;
                    }
                    case 7: {
                        this.readCOMPOSITE();
                        continue;
                    }
                    default: {
                        throw new RuntimeException("Corrupted data file");
                    }
                }
            }
            this.bb = null;
            return this;
        }
        catch (IOException ex) {
            ex.printStackTrace();
            return null;
        }
    }
    
    static {
        CharsetMapping.comparatorBytes = new Comparator<CharsetMapping.Entry>() {
            @Override
            public int compare(CharsetMapping.Entry entry, CharsetMapping.Entry entry2) {
                return entry.bs - entry2.bs;
            }
            
            @Override
            public boolean equals(Object o) {
                return this == o;
            }
        };
        CharsetMapping.comparatorCP = new Comparator<CharsetMapping.Entry>() {
            @Override
            public int compare(CharsetMapping.Entry entry, CharsetMapping.Entry entry2) {
                return entry.cp - entry2.cp;
            }
            
            @Override
            public boolean equals(Object o) {
                return this == o;
            }
        };
        CharsetMapping.comparatorComp = new Comparator<CharsetMapping.Entry>() {
            @Override
            public int compare(CharsetMapping.Entry entry, CharsetMapping.Entry entry2) {
                int n = entry.cp - entry2.cp;
                if (n == 0) {
                    n = entry.cp2 - entry2.cp2;
                }
                return n;
            }
            
            @Override
            public boolean equals(Object o) {
                return this == o;
            }
        };
    }
    
    public static class Entry
    {
        public int bs;
        public int cp;
        public int cp2;
    }
}
